@@ -17,27 +17,38 @@
 #include <caml/fail.h>
 #include <caml/hash.h>
 #include <caml/unixsupport.h>
+#include <caml/threads.h>
 
 #include <ffi.h>
 
-#include "../ctypes/managed_buffer_stubs.h"
-#include "../ctypes/type_info_stubs.h"
-#include "../ctypes/raw_pointer.h"
+#include "../ctypes/ctypes_managed_buffer_stubs.h"
+#include "../ctypes/ctypes_type_info_stubs.h"
+#include "../ctypes/ctypes_raw_pointer.h"
 
 /* TODO: support callbacks that raise exceptions?  e.g. using
    caml_callback_exn etc.  */
 
-/* TODO: thread support */
+/* Register a C thread with the OCaml runtime.  By default this simply
+   fails.  The ctypes.foreign.threaded subpackage overrides it to call
+   [caml_c_thread_register].
+ */
+static int ctypes_thread_register_fail(void)
+{
+  caml_failwith("ctypes_thread_register unavailable: "
+		"please link with the threads library");
+}
+int (*ctypes_thread_register)(void) = ctypes_thread_register_fail;
+
 
 /* An OCaml function that converts resolves identifiers to OCaml functions */
 static value retrieve_closure_;
 
 /* Resolve identifiers to OCaml functions */
-static value retrieve_closure(int key)
+static value retrieve_closure(intnat key)
 {
   CAMLparam0 ();
   CAMLlocal1(result);
-  result = caml_callback_exn(retrieve_closure_, Val_int(key));
+  result = caml_callback_exn(retrieve_closure_, Val_long(key));
 
   if (Is_exception_result(result)) {
     caml_raise_constant(*caml_named_value("CallToExpiredClosure"));
@@ -106,7 +117,7 @@ static struct callspec {
   /* The maximum element alignment */
   size_t max_align;
 
-  /* The state of the bufferspec value. */
+  /* The state of the callspec value. */
   enum { BUILDING, CALLSPEC } state;
 
   /* A null-terminated array of size `nelements' types */
@@ -115,6 +126,25 @@ static struct callspec {
   /* return value offset */
   size_t roffset;
 
+  /* return offset adjustment.
+
+     libffi promotes return types that are less than the size of the
+     system register to the word-sized type ffi_arg.  On a big-endian
+     system this means that the address where libffi writes the return
+     value is not always the same as the address from which ctypes
+     should read the value.
+  */
+  size_t radjustment;
+
+
+  /* The context in which the call should run: whether errno is
+     checked, whether the runtime lock is released, and so on. */
+  struct call_context {
+    int check_errno:1;
+    int runtime_lock:1;
+    int thread_registration:1;
+  } context;
+
   /* The libffi call interface structure.  It would be nice for this member to
      be a value rather than a pointer (to save a layer of indirection) but the
      ffi_closure structure keeps the address of the structure, and the GC can
@@ -123,25 +153,26 @@ static struct callspec {
   ffi_cif *cif;
 
 } callspec_prototype = {
-  0, 0, 0, 0, BUILDING, NULL, -1, NULL
+  0, 0, 0, 0, BUILDING, NULL, -1, 0, { 0, 0 }, NULL
 };
 
 
-static void finalize_bufferspec(value v)
+static void finalize_callspec(value v)
 {
   struct callspec *callspec = Data_custom_val(v);
-  free(callspec->args);
-  free(callspec->cif);
+  caml_stat_free(callspec->args);
+  caml_stat_free(callspec->cif);
 }
 
 
 static struct custom_operations callspec_custom_ops = {
-  "ocaml-ctypes:bufferspec",
-  finalize_bufferspec,
+  "ocaml-ctypes:callspec",
+  finalize_callspec,
   custom_compare_default,
   custom_hash_default,
   custom_serialize_default,
-  custom_deserialize_default
+  custom_deserialize_default,
+  custom_compare_ext_default
 };
 
 /* We store two things in the callbuffer: a "scratch" area for passing
@@ -187,13 +218,21 @@ static void populate_arg_array(struct callspec *callspec,
 
 
 /* Allocate a new C call specification */
-/* allocate_callspec : unit -> callspec */
-value ctypes_allocate_callspec(value unit)
+/* allocate_callspec : check_errno:bool -> runtime_lock:bool -> callspec */
+value ctypes_allocate_callspec(value check_errno, value runtime_lock,
+                               value thread_registration)
 {
+  struct call_context context = {
+    Int_val(check_errno),
+    Int_val(runtime_lock),
+    Int_val(thread_registration),
+  };
+
   value block = caml_alloc_custom(&callspec_custom_ops,
                                   sizeof(struct callspec), 0, 1);
-  memcpy(Data_custom_val(block), &callspec_prototype,
-         sizeof(struct callspec));
+  struct callspec *spec = Data_custom_val(block);
+  memcpy(spec, &callspec_prototype, sizeof(struct callspec));
+  spec->context = context;
   return block;
 }
 
@@ -229,7 +268,31 @@ value ctypes_add_argument(value callspec_, value argument_)
   callspec->max_align = argtype->alignment > callspec->max_align
     ? argtype->alignment
     : callspec->max_align;
-  CAMLreturn(Val_int(offset));
+  CAMLreturn(Val_long(offset));
+}
+
+static int ffi_return_type_adjustment(ffi_type *f)
+{
+#ifdef ARCH_BIG_ENDIAN
+  /* An adjustment is needed (on bigendian systems) for integer types
+     less than the size of a word */
+  if (f->size < sizeof(ffi_arg)) {
+    switch (f->type) {
+    case FFI_TYPE_INT:
+    case FFI_TYPE_UINT8:
+    case FFI_TYPE_SINT8:
+    case FFI_TYPE_UINT16:
+    case FFI_TYPE_SINT16:
+    case FFI_TYPE_UINT32:
+    case FFI_TYPE_SINT32:
+    case FFI_TYPE_UINT64:
+    case FFI_TYPE_SINT64:
+      return sizeof(ffi_arg) - f->size;
+    default: break;
+    }
+  }
+#endif
+  return 0;
 }
 
 
@@ -249,9 +312,11 @@ value ctypes_prep_callspec(value callspec_, value abi_, value rtype)
   /* Add the (aligned) space needed for the return value */
   callspec->roffset = aligned_offset(callspec->bytes,
                                      rffitype->alignment);
+  callspec->radjustment = ffi_return_type_adjustment(rffitype);
   callspec->bytes = callspec->roffset + rffitype->size;
 
-  /* Allocate an extra word after the return value space to work
+
+  /* Allocate an extra word after the return value space, to work
      around a bug in libffi which causes it to write past the return
      value space.
 
@@ -275,16 +340,19 @@ value ctypes_prep_callspec(value callspec_, value abi_, value rtype)
 
 /* Call the function specified by `callspec', passing arguments and return
    values in `buffer' */
-/* call : raw_pointer -> callspec -> (raw_pointer -> Obj.t array -> unit) ->
-          (raw_pointer -> 'a) -> 'a */
-value ctypes_call(value function, value callspec_, value argwriter, value rvreader)
+/* call : string -> _ fn Fat.t -> callspec ->
+          (raw_pointer -> Obj.t array -> unit) -> (raw_pointer -> 'a) -> 'a */
+value ctypes_call(value fnname, value function, value callspec_,
+                  value argwriter, value rvreader)
 {
-  CAMLparam4(function, callspec_, argwriter, rvreader);
+  CAMLparam5(fnname, function, callspec_, argwriter, rvreader);
   CAMLlocal3(callback_arg_buf, callback_val_arr, callback_rv_buf);
 
   struct callspec *callspec = Data_custom_val(callspec_);
   int roffset = callspec->roffset;
+  struct call_context context = callspec->context;
   size_t nelements = callspec->nelements;
+  ffi_cif *cif = callspec->cif;
 
   assert(callspec->state == CALLSPEC);
 
@@ -292,7 +360,8 @@ value ctypes_call(value function, value callspec_, value argwriter, value rvread
   size_t bytes = compute_arg_buffer_size(callspec, &arg_array_offset);
 
   char *callbuffer = alloca(bytes);
-  char *return_slot = callbuffer + roffset;
+  char *return_write_slot = callbuffer + roffset;
+  char *return_read_slot = return_write_slot + callspec->radjustment;
 
   populate_arg_array(callspec, (struct callbuffer *)callbuffer,
                      (void **)(callbuffer + arg_array_offset));
@@ -303,7 +372,7 @@ value ctypes_call(value function, value callspec_, value argwriter, value rvread
 
   void **val_refs = alloca(sizeof(void*) * nelements);
 
-  int arg_idx;
+  unsigned arg_idx;
   for(arg_idx = 0; arg_idx < Wosize_val(callback_val_arr); arg_idx++) {
     value arg_tuple = Field(callback_val_arr, arg_idx);
     /* <4.02 initialize to 0; >=4.02 initialize to unit. */
@@ -314,81 +383,194 @@ value ctypes_call(value function, value callspec_, value argwriter, value rvread
 
     /* Only strings have defined semantics for now. */
     assert(Is_block(arg_ptr) && Tag_val(arg_ptr) == String_tag);
-    val_refs[arg_idx] = String_val(arg_ptr) + Int_val(arg_offset);
+    val_refs[arg_idx] = String_val(arg_ptr) + Long_val(arg_offset);
 
     ((void**)(callbuffer + arg_array_offset))[arg_idx] = &val_refs[arg_idx];
   }
 
-  void (*cfunction)(void) = (void (*)(void)) CTYPES_TO_PTR(function);
+  void (*cfunction)(void) = (void (*)(void)) CTYPES_ADDR_OF_FATPTR(function);
+  int check_errno = context.check_errno;
+  int saved_errno = 0;
+  if (context.runtime_lock)
+  {
+    caml_release_runtime_system();
+  }
 
-  ffi_call(((struct callspec *)Data_custom_val(callspec_))->cif,
+  if (check_errno)
+  {
+    errno=0;
+  }
+
+  ffi_call(cif,
            cfunction,
-           return_slot,
+           return_write_slot,
            (void **)(callbuffer + arg_array_offset));
+  if (check_errno)
+  {
+    saved_errno=errno;
+  }
 
-  callback_rv_buf = CTYPES_FROM_PTR(return_slot);
-  CAMLreturn(caml_callback(rvreader, callback_rv_buf));
-}
+  if (context.runtime_lock)
+  {
+    caml_acquire_runtime_system();
+  }
 
-
-/* call_errno : string -> raw_pointer -> callspec ->
-               (raw_pointer -> unit) ->
-               (raw_pointer -> 'a) -> 'a */
-value ctypes_call_errno(value fnname, value function, value callspec_,
-                        value argwriter, value rvreader)
-{
-  CAMLparam5(fnname, function, callspec_, argwriter, rvreader);
-
-  errno = 0;
-  CAMLlocal1(rv);
-  rv = ctypes_call(function, callspec_, argwriter, rvreader);
-  if (errno != 0)
+  if (check_errno && saved_errno != 0)
   {
     char *buffer = alloca(caml_string_length(fnname) + 1);
     strcpy(buffer, String_val(fnname));
-    unix_error(errno, buffer, Nothing);
+    unix_error(saved_errno, buffer, Nothing);
   }
-  CAMLreturn(rv);
+
+  callback_rv_buf = CTYPES_FROM_PTR(return_read_slot);
+  CAMLreturn(caml_callback(rvreader, callback_rv_buf));
 }
 
 typedef struct closure closure;
 struct closure
 {
-  ffi_closure closure;
-  int         fnkey;
+  ffi_closure         closure;
+  intnat                fnkey;
+  struct call_context context;
+  void*                 fnptr;
 };
 
 enum boxedfn_tags { Done, Fn };
 
+/* callback_handler_with_lock must only be called while the runtime lock is
+   held. */
+static void callback_handler_with_lock(ffi_cif *cif,
+                                       void *ret,
+                                       void **args,
+                                       void *user_data)
+{
+  CAMLparam0 ();
+
+  CAMLlocal2(boxedfn, argptr);
+  closure *closure = user_data;
+
+  boxedfn = retrieve_closure(closure->fnkey);
+
+  int i, arity = cif->nargs;
+
+  switch (arity)
+  {
+    case 0:
+    {
+      assert (Tag_val(boxedfn) == Fn);
+      boxedfn = caml_callback(Field(boxedfn, 0), Val_unit);
+      break;
+    }
+    default:
+    {
+      for (i = 0; i < arity; i++)
+      {
+        void *cvalue = args[i];
+        assert (Tag_val(boxedfn) == Fn);
+        /* unbox and call */
+        argptr = CTYPES_FROM_PTR(cvalue);
+        boxedfn = caml_callback(Field(boxedfn, 0), argptr);
+      }
+      break;
+    }
+  }
+
+  /* now store the return value */
+  assert (Tag_val(boxedfn) == Done);
+
+  argptr = CTYPES_FROM_PTR(ret);
+  caml_callback(Field(boxedfn, 0), argptr);
+
+  /* workaround for libffi api: small integers must be promoted to
+   * full word size (sign/zero extended) */
+  if (cif->rtype->size < sizeof(ffi_arg)) {
+    int do_nothing = 0;
+    ffi_arg x;
+    switch (cif->rtype->type) {
+    case FFI_TYPE_INT:
+      x = *(int*)ret;
+      break;
+    case FFI_TYPE_UINT8:
+      x = *(uint8_t*)ret;
+      break;
+    case FFI_TYPE_SINT8:
+      x = *(int8_t*)ret;
+      break;
+    case FFI_TYPE_UINT16:
+      x = *(uint16_t*)ret;
+      break;
+    case FFI_TYPE_SINT16:
+      x = *(int16_t*)ret;
+      break;
+    case FFI_TYPE_UINT32:
+      x = *(uint32_t*)ret;
+      break;
+    case FFI_TYPE_SINT32:
+      x = *(int32_t*)ret;
+      break;
+    case FFI_TYPE_UINT64:
+      x = *(uint64_t*)ret;
+      break;
+    case FFI_TYPE_SINT64:
+      x = *(int64_t*)ret;
+      break;
+    default:
+      do_nothing = 1;
+      break;
+    }
+    if ( do_nothing == 0 ) {
+      *(ffi_arg*)ret = x;
+    }
+  }
+  CAMLreturn0;
+}
+
+
 static void callback_handler(ffi_cif *cif,
                              void *ret,
                              void **args,
                              void *user_data)
 {
-  CAMLparam0 ();
+  closure *closure = user_data;
 
-  CAMLlocal2(boxedfn, argptr);
-  boxedfn = retrieve_closure(*(int *)user_data);
-
-  int i, arity = cif->nargs;
-  for (i = 0; i < arity; i++)
+  if (closure->context.thread_registration)
   {
-    void *cvalue = args[i];
-    assert (Tag_val(boxedfn) == Fn);
-    /* unbox and call */
-    argptr = CTYPES_FROM_PTR(cvalue);
-    boxedfn = caml_callback(Field(boxedfn, 0), argptr);
+    ctypes_thread_register();
   }
 
-  /* now store the return value */
-  assert (Tag_val(boxedfn) == Done);
-  argptr = CTYPES_FROM_PTR(ret);
-  caml_callback(Field(boxedfn, 0), argptr);
+  if (closure->context.runtime_lock)
+  {
+    caml_acquire_runtime_system();
+  }
 
-  CAMLreturn0;
+  callback_handler_with_lock(cif, ret, args, user_data);
+
+  if (closure->context.runtime_lock)
+  {
+    caml_release_runtime_system();
+  }
 }
 
 
+static void finalize_closure(value v)
+{
+  struct closure **closure = Data_custom_val(v);
+  ffi_closure_free(*closure);
+}
+
+/* A custom object whose purpose is the finalizer that calls
+   ffi_closure_free */
+static struct custom_operations closure_custom_ops = {
+  "ocaml-ctypes:closure",
+  finalize_closure,
+  custom_compare_default,
+  custom_hash_default,
+  custom_serialize_default,
+  custom_deserialize_default,
+  custom_compare_ext_default
+};
+
+
 /* Construct a pointer to an OCaml function represented by an identifier */
 /* make_function_pointer : callspec -> int -> raw_pointer */
 value ctypes_make_function_pointer(value callspec_, value fnid)
@@ -401,27 +583,34 @@ value ctypes_make_function_pointer(value callspec_, value fnid)
 
   void (*code_address)(void) = NULL;
 
-  /* TODO: we need to call ffi_closure_free at some point.  This function
-     should return a managed object to which we can attach a finaliser for the
-     closure.
-  */
   closure *closure = ffi_closure_alloc(sizeof *closure, (void *)&code_address);
 
   if (closure == NULL) {
     caml_raise_out_of_memory();
   } else {
-    closure->fnkey = Int_val(fnid);
+    closure->fnkey = Long_val(fnid);
+    closure->context = callspec->context;
+    closure->fnptr = code_address;
 
     ffi_status status =  ffi_prep_closure_loc
       ((ffi_closure *)closure,
        callspec->cif,
        callback_handler,
-       &closure->fnkey,
+       closure,
        (void *)code_address);
 
     ctypes_check_ffi_status(status);
 
-    codeptr = CTYPES_FROM_PTR((void *)code_address);
+    codeptr =
+      caml_alloc_custom(&closure_custom_ops, sizeof(struct closure *), 1, 1);
+    *(struct closure **)Data_custom_val(codeptr) = closure;
+
     CAMLreturn (codeptr);
   }
 }
+
+/* Extract the raw address from a function pointer object */ 
+value ctypes_raw_address_of_function_pointer(value closure)
+{
+  return CTYPES_FROM_PTR((*(struct closure **)Data_custom_val(closure))->fnptr);
+}
