@@ -15,7 +15,8 @@ struct
 
   let charish = view ~read:Char.chr ~write:Char.code int
 
-  let nullable_intptr = Foreign.funptr_opt (int @-> int @-> returning int)
+  let nullable_intptr = Foreign.funptr_opt Ctypes.(int @-> int @->
+                                                   returning int)
 
   let concat_strings = foreign "concat_strings"
     (ptr string @-> int @-> ptr char @-> returning void)
@@ -27,4 +28,7 @@ struct
     (int @-> returning nullable_intptr)
   let accepting_possibly_null_funptr = foreign "accepting_possibly_null_funptr"
     (nullable_intptr @-> int @-> int @-> returning int)
+
+  let strcmp = foreign "strcmp"
+    (string @-> string @-> returning int)
 end
